(in-package :6502)

;;;; Building a mini-language on top of the assembler..

;; TODO: Instead of playing stupid macro games with gensym everywhere to
;; maintain hygiene, I ought to establish an idiom for binding a local
;; assembly context which delegates code emission but uses a local symbol
;; table, binding the caller's context back into place when compiling
;; user code (so that the user's namespace comes back into scope.)

;; TODO: We ought to have a protocol for handler what assemble-if currently
;; treats as a 'dumb' closure to call to generate each half of the conditional.
;; Specifically, we could treat that as an arbitrary object with a GF to
;; invoke generation, for which the method on the 'function' class would
;; simply use funcall. Why? It would let us communicate and optimize certain
;; properties, such as that the body of that block jumps (so we don't need
;; to emit the connecting jump).

;; Assemble an if-then-else construct. The 'branch-compiler' is invoked
;; to generate conditional branch to the else clause. If the 'else-compiler'
;; is omitted, the jump following the "then" clause will be optimized away.
(defun assemble-if (branch-compiler then-compiler &optional else-compiler)
  (let ((else-sym    (gensym "ELSE"))
        (finally-sym (gensym "FINALLY")))
    (assembling nil ()
      (funcall branch-compiler else-sym)
      (funcall then-compiler)
      (when else-compiler (jmp finally-sym))
      (set-label else-sym)
      (when else-compiler (funcall else-compiler))
      (set-label finally-sym))))

(defgeneric condition-to-branch! (condition)
  (:documentation "Produce a closure capable of generating a branch to the given argument
if the condition is not true." ))
(defmethod condition-to-branch! ((condition (eql :positive)))    #'bmi)
(defmethod condition-to-branch! ((condition (eql :negative)))    #'bpl)
(defmethod condition-to-branch! ((condition (eql :carry)))       #'bcc)
(defmethod condition-to-branch! ((condition (eql :no-carry)))    #'bcs)
(defmethod condition-to-branch! ((condition (eql :zero)))        #'bne)
(defmethod condition-to-branch! ((condition (eql :not-zero)))    #'beq)
(defmethod condition-to-branch! ((condition (eql :equal)))       #'bne)
(defmethod condition-to-branch! ((condition (eql :not-equal)))   #'beq)
(defmethod condition-to-branch! ((condition (eql :overflow)))    #'bvc)
(defmethod condition-to-branch! ((condition (eql :no-overflow))) #'bvs)

(defmacro xif (condition then &optional else)
  `(assemble-if
    ,(condition-to-branch! condition)
    (lambda () ,@then)
    ,(and else `(lambda () ,@else))))

;; Test of 'xif' construct:
#+nil
(dumpbin "/tmp/test.bin"
  (assembling nil (basic-assembly-context :origin #x8000)
    (nop)
    (lda (mem #x2002))
    (xif :positive
        ((sta (mem #x600))
         (lda 0))
        ((sta (zp #x30))
         (lda 1)))
    (clc)))



         
  
    