(in-package :6502) ;; Dammit, we need our own package

(defvar *asset-directory* "/home/hefner/cl/6502/")

(defun loadbin (filename)
  (with-open-file (in filename :element-type '(unsigned-byte 8))
    (let ((data (make-array (file-length in))))
      (read-sequence data in)
      data)))

(defun load-asset (name) (loadbin (merge-pathnames name *asset-directory*)))

;;; 6502 and NES Macros

(defun poke! (address value)
  (lda value)
  (sta (cond ((and (integerp address) (< address #x100)) (zp address))
             ((integerp address) (mem address))
             (t address))))

(defgeneric address-of (x))
(defmethod address-of ((expr absolute-mode)) (parameter expr))
(defmethod address-of ((address integer)) address)
(defmethod address-of ((sym symbol)) (fixup (address (get-label sym)) address))

(defun poke16! (address value)          ; FIXME: Forward references!
  (let ((address (address-of address)))
    (poke! (make-fixup address #'identity) (imm (fixup (x value) (lsb-of x))))
    (poke! (make-fixup address #'1+)       (imm (fixup (x value) (msb-of x))))))


(defun set-vram-addr! (address)
  (poke! +vram-addr+ (msb-of address))
  (poke! +vram-addr+ (lsb-of address)))

;; Indirect call macro
(defun call-indirect! (dest)
  (let ((post-sym (gensym "Indirect call return point")))
    (lda (label-msb post-sym))
    (pha)
    (lda (label-lsb post-sym))
    (pha)
    (jmp (indirect dest))
    (set-label post-sym)))
    

;;; Demo1 Program

(let* (#| Zero-Page Variables |#
       (tint-shifter   (zp #x01))       ; Rotating color emphasis bit
       (frame-count    (zp #x02))       ; Inner loop counter (decrement per frame)
       (hscroll-offset (zp #x04))       ; Current scroll offset
       (music-counter  (zp #x30))       ; Frames until next music event
       (channel-1-idx  (zp #x31))       ; Index/state variable for first sound channel

       #| Other Variables |#
       (dw-music-proc  (mem #x200)) ; Hah..
       #|*** Program Code ***|#
       (program
 (assembling nil (basic-assembly-context :origin #x8000)
   ;;;; Vertical Blank Handler
   (set-label 'vblank-handler)
   (rti)
   
   ;;;; BRK/IRQ Handler
   (set-label 'irq-handler)
   (rti)
   
   ;;;; Entry Point
   (set-label 'entry-point)
   (sei)                                ; Disable interrupts
   (lda #xFF)                           ; Initialize stack pointer to #xFF
   (tax)
   (txs)
   (cld)                                ; Clear decimal flag
   (poke! #x4017 #b01000000)            ; NTSC mode, disable 2A03 interrupts
   (cli)                                ; Enable interrupts

   ;; Disable video, so we can write to VRAM freely (is this true? :)
   (poke! +ppu-cr1+ #b00000000)
   (poke! +ppu-cr2+ #b00000000)

   ;; Write a palette
   (set-vram-addr! #x3F00)
   (let ((palette '(#x00 #x10 #x20 #x30  #x00 #x12 #x22 #x32  #x00 #x11 #x21 #x31  #x00 #x19 #x29 #x39
                    #x00 #x13 #x23 #x33  #x00 #x14 #x24 #x34  #x00 #x15 #x25 #x35  #x00 #x16 #x26 #x36)))

     (dolist (color palette) (poke! +vram-io+ color)))

   ;; Fill name and attribute table
   (set-vram-addr! #x2020)

   (let* ((num-loops (* 32 28))
          (outer-loops (1+ (ldb (byte 8 8) num-loops)))
          (loop-modulus (ldb (byte 8 0) num-loops)))
     (ldy outer-loops)
     (ldx loop-modulus)
     (lda 0)

     ;; This is dumb, :loop is not local to this block. Need a named block macro.
     (assembling :loop ()
       (sta (mem +vram-io+))
       (clc)
       (adc 1)
       (dex)
       (bne :loop)
       (dey)
       (bne :loop)))

   ;; Enable video
   (lda (mem +ppu-status+))
   (poke! +vram-addr+ #x0000)
   (poke! +ppu-cr1+ #b10000000)
   (poke! +ppu-cr2+ #b00011110)

   (nop)
   (nop)
   (nop)
   (nop)
   
   ;(jmp (mem *origin*))                 ; Spin infinitely..   

   (jsr 'music-init)                    ; Initialize music player
   
   (set-label :cycle-frame)
   (lda 60)                            ; Cycle every N frames
   (sta frame-count)

   ;(jsr 'sound-test)
   
   (set-label :frame-loop)

   ;; Before entering vertical blanking period, let the music player run.
   (format *trace-output* "~&Music player call lives at ~X~%" *origin*)
   (dec music-counter)                  ; Decrement current cycle count
   (xif :zero ((call-indirect! dw-music-proc)))
   (nop)
   (nop)
   (nop)
   (nop)

   (set-label :vblank-wait)             ; Wait for VBLANK
   (lda (mem +ppu-status+))
   (bpl :vblank-wait)

   (poke! +ppu-cr2+ #b11111110)         ; Sanity check, make sure we are updating every frame

   (poke! +spr-address+ #x00)           ; Program sprite 0
   (poke! +spr-io+ #x40)
   (poke! +spr-io+ #x10)
   (poke! +spr-io+ #x01)
   (poke! +spr-io+ #x40)

   (poke! +vram-scroll+ 0)              ; Just for fun, fuck with the scroll registers
   (poke! +vram-scroll+ 0)

   (lda tint-shifter)
   (ora #b00011110)
   (sta (mem +ppu-cr2+))

   (lda (mem +ppu-status+))   
   (jsr :sprite0-hit-loop)   

   (dec frame-count)
   (bne :frame-loop)

   (lda tint-shifter)
   (asl)

   (xif :equal
        ((ora #b00100000)))
;   (bne :foo)
;   (ora #b00100000)
;   (set-label :foo)
   (sta tint-shifter)

   (jmp :cycle-frame)

   (set-label :sprite0-hit-loop)        ; Wait for sprite0 hit  
   (dotimes (i 3) (jsr :stupid-waitloop)) ;; Magic voodoo delay, otherwise hit flag a misbehaves?
   (set-label :sprite0-hit-loop-1)
   (lda (mem +ppu-status+))
   (anda #b01000000)   
   (beq :sprite0-hit-loop-1)

   (poke! +vram-scroll+ hscroll-offset)
   (poke! +vram-scroll+ 0)
   (inc hscroll-offset)
   (poke! +ppu-cr2+ #b00011110)
;   (dotimes (i 4) (jsr :stupid-waitloop))
;   (poke! +ppu-cr2+ #b10011110) ;; Hmm.. This ALMOST works.. WTF?
   (rts)

   (nop)
   (nop)
   (nop)

   (set-label :stupid-waitloop)
   (ldx 255)
   (set-label :stupid-waitloop-1)
   (dex)
   (bne :stupid-waitloop-1)
   (rts)

   

   ;;;; Sound test
   (assembling sound-test ()
     (poke! +papu-control+ #b000000011) ; Enable square channel 1

     (poke! +pulse1-control+ #b10000010)
     (poke! +pulse1-ramp+ 0)
     (poke! +pulse1-fine+ #x80)
     (poke! +pulse1-coarse+ #b11111000)

     (poke! +pulse2-control+ #b01000010)
     (poke! +pulse2-ramp+ 0)
     (poke! +pulse2-fine+ #x81)
     (poke! +pulse2-coarse+ #b11111000)
     (rts))

;   (flet ((note (n) (/ 1790000 (* 440 (expt 2 n)) 16))
;          (n (note &optional (octave 0)) (+ (* octave 12) n))) )
   
   ;;;; Music Player
   (let ((tempo 10))
     (assembling music-init ()
       (poke! music-counter tempo)
       (hef:debugf "music-init" *origin*)
       (poke16! dw-music-proc (fixup (x (get-label 'music-run)) x)) ; YUCK.
       (rts))

     (assembling music-run ()
       (jsr 'sound-test)
       (poke! music-counter tempo)
       (rts)))

   ;;;; Interrupt Vectors
   (advance-to +6502-nmi-vector+)
   (dw 'vblank-handler)
   (dw 'entry-point)
   (dw 'irq-handler))))
   
  (dumpbin "/tmp/demo1.nes"
           (concatenate 'vector
                        (ines-header 2 1)
                        program
                        (load-asset "smb1.chr")))
  (values))