(in-package :6502)  ;; FIXME: Export appropriate symbols from 6502 and put this in its own package.

;;;; NES Hardware Registers

(defconstant +ppu-cr1+ #x2000)
(defconstant +ppu-cr2+ #x2001)
(defconstant +ppu-status+ #x2002)
(defconstant +spr-address+ #x2003)
(defconstant +spr-io+ #x2004)
(defconstant +vram-scroll+ #x2005)
(defconstant +vram-addr+ #x2006)
(defconstant +vram-io+ #x2007)

(defconstant +pulse1-control+ #x4000)
(defconstant +pulse1-ramp+ #x4001)
(defconstant +pulse1-fine+ #x4002)
(defconstant +pulse1-coarse+ #x4003)
(defconstant +pulse2-control+ #x4004)
(defconstant +pulse2-ramp+ #x4005)
(defconstant +pulse2-fine+ #x4006)
(defconstant +pulse2-coarse+ #x4007)
(defconstant +tri-cr1+ #x4008)
(defconstant +tri-cr2+ #x4009)
(defconstant +tri-freq1+ #x400A)
(defconstant +tri-freq2+ #x400B)
(defconstant +noise-control+ #x400C)
(defconstant +noise-freq1+ #x400E)
(defconstant +noise-freq2+ #x400F)
(defconstant +dmod-control+ #x4010)
(defconstant +dmod-da+ #x4011)
(defconstant +dmod-address+ #x4012)
(defconstant +dmod-data-length+ #x4013)

(defconstant +sprite-dma-control+ #x4014)
(defconstant +papu-control+ #x4015)

(defconstant +joypad-1+ #x4016)
(defconstant +joypad-2+ #x4017)

;;;; ROM Output Utilities

(defconstant +ines-horizontal-mirroring+ 0)
(defconstant +ines-vertical-mirroring+ 1)
(defconstant +ines-four-screen+ 8)

(defun ines-header (prg-pages chr-pages &key
                              (mirror-mode +ines-horizontal-mirroring+)
                               (mapper 0) trainerp sram)
  (let ((control-1 (logior mirror-mode
                           (ash (logand mapper #x0F) 4)
                           (if sram 2 0)
                           (if trainerp 4 0)))
        (control-2 (logand mapper #xF0)))
    (vector #x4E #x45 #x53 #x1A prg-pages chr-pages control-1 control-2 0 0 0 0 0 0 0 0)))

;;;; Test Program!

(defun assemble-test-program ()
  (assembling nil (basic-assembly-context :origin #x8000)
    ;;;; Program Entry Point
    (set-label 'entry-point)
    ;; Configure PPU
    (lda #b10000000)
    (sta (mem +ppu-cr1+))
    (lda #b00000100)
    (sta (mem +ppu-cr2+))
    ;; Spin indefinitely, remaining work is done in VBlank handler.
    (jmp (mem *origin*))

    ;;;; VBlank Handler
    (set-label 'vblank-handler)
    ;;; Write to palette register 0 (#x3F00 in VRAM):
    (lda #x3F) ; Program address port - high byte
    (sta (mem +vram-addr+))
    (lda #x00)    ; low byte
    (sta (mem +vram-addr+))
    (lda (zp 0)) ; Load color counter variable
    (sta (mem +vram-io+)) ; Write to VRAM IO Port
    (inc (zp 0)) ; Increment color counter
    (rti)

    ;;;; IRQ/Break Handler
    (set-label 'brk-handler)
    (rti)

    ;; Interrupt Vectors
    (advance-to +6502-nmi-vector+)
    (dw 'vblank-handler) ;; NMI
    (dw 'entry-point)    ;; RESET
    (dw 'brk-handler)))  ;; BRK/IRQ

(defun dump-test-rom (&optional (filename "/tmp/test.nes"))
  (let* ((program (assemble-test-program))
         (characters (make-array #x2000 :initial-element #x5A)))
    (assert (zerop (mod (length program) #x4000)))
    (assert (zerop (mod (length characters) #x2000)))
    (dumpbin filename
             (concatenate 'vector
                          (ines-header (/ (length program) #x4000)
                                       (/ (length characters) #x2000))
                          program
                          characters))))


